#include <stdlib.h>
#include <math.h>
#include <R.h>
#include <Rmath.h>        /* for 'runif' */
#include <R_ext/Utils.h>  /* for rsort_with_index */


/* Modified from entry in: 'R-2.4.0/include/R_ext/RS.h' 
#define Calloc(n, t)   (t *) calloc( (size_t) (n), sizeof(t) )
*/


/* Replace later(?) with boolean typed variables? */
/* Need to #include<stdbool.h> for that */
#define TRUE 1
#define FALSE 0


/* =========================================================================== */

int Sign(double x)
{
    return ((x > 0) ? 1 : ((x == 0)? 0 : -1));
}


/* =========================================================================== */

/* This function was taken from R-2.4.0/src/main/sort.c, modified  */
/* by replacing "rcmp(x[j - h],v,TRUE) > 0" with "(x[j - h] > v)"  */
/* then renamed (to avoid getting the R function when compiling) */

void indexSort(double *x, int *indx, int n)
{
    double v;
    int i, j, h, iv;

    for (h = 1; h <= n / 9; h = 3 * h + 1);
    for (; h > 0; h /= 3)
        for (i = h; i < n; i++) {
            v = x[i]; iv = indx[i];
            j = i;
            while (j >= h && (x[j - h] > v))
                 { x[j] = x[j - h]; indx[j] = indx[j-h]; j -= h; }
            x[j] = v; indx[j] = iv;
        }
}


/* =========================================================================== */

/* Tabulation functions */


void freq1D(int *x, int lev, int N, double *freqVec) 
{
    int i; 

    /* initialize region of tabVec that we'll be using to zero */
    for (i = 0; i < lev; i++)
        freqVec[i] = 0;

    /* level codes start at 1, so need to subtract 1 to start them at 0 */
    for (i = 0; i < N; i++)
        freqVec[ x[i]-1 ]++ ; 

    /* convert counts to freqs/probabilities */ 
    for (i = 0; i < lev; i++)
        freqVec[i] = freqVec[i]/N; 
}


void freq2D(int *x, int *y, int x_lev, int y_lev, int N, double **freqMat) 
{
    int i, j; 

    /* initialize the region of freqMat that we'll be using to zero */
    for (i = 0; i < x_lev; i++)
        for (j = 0; j < y_lev; j++)
            freqMat[i][j] = 0;

    /* level codes start at 1, so need to subtract 1 to start them at 0 */
    for (i = 0; i < N; i++)
        freqMat[ x[i]-1 ][ y[i]-1 ]++ ; 

    /* convert counts to freqs/probabilities */ 
    for (i = 0; i < x_lev; i++)
        for (j = 0; j < y_lev; j++)
            freqMat[i][j] = freqMat[i][j]/N;
}


void freq3D(int *x, int *y, int *z, int x_lev, int y_lev, int z_lev,
            int N, double ***freqArray) 
{
    int i, j, k; 

    /* initialize the region of freqMat that we'll be using to zero */
    for (i = 0; i < x_lev; i++)
        for (j = 0; j < y_lev; j++)
            for (k = 0; k < z_lev; k++)
                freqArray[i][j][k] = 0;

    /* level codes start at 1, so need to subtract 1 to start them at 0 */
    for (i = 0; i < N; i++)
        freqArray[ x[i]-1 ][ y[i]-1 ][ z[i]-1 ]++ ; 

    /* convert counts to freqs/probabilities */ 
    for (i = 0; i < x_lev; i++)
        for (j = 0; j < y_lev; j++)
            for (k = 0; k < z_lev; k++)
                freqArray[i][j][k] = freqArray[i][j][k]/N;
}



/* =========================================================================== */


void calcStatsTriples(int **x, int Nseq, int Npos, int Ntriples, 
                      int *posI, int *posJ, int *posK, int *Nlev, int maxLvl,
                      double **jntprobIJ, double **jntprobJK, 
                      double **jntprobIK, double ***jntprobIJK, double **p,
                      double *sum_PlogP, double *M, double *Mstar, int *anyNAN)
{
    int triple;
    int pos_i, pos_j, pos_k, i_lev, j_lev, k_lev;
    double pIJK, prod1Dmarg, prod2Dmarg, norm_const;
    long double tmpVal;

    for (triple = 0; triple < Ntriples; triple++)
    {
        /* Note: the minus one is to adjust for vector indexing */
        pos_i = posI[triple] - 1;
        pos_j = posJ[triple] - 1;
        pos_k = posK[triple] - 1;


        /* calc. joint frequencies of positions i and j */
        if (triple > 0) 
        {
            /* If we have same i,j values as last triplet, don't recompute */
            if (pos_i != posI[triple - 1]  ||  pos_j != posJ[triple - 1])
                freq2D(x[pos_i], x[pos_j], Nlev[pos_i], Nlev[pos_j],
                       Nseq, jntprobIJ);
        } else
        {
            freq2D(x[pos_i], x[pos_j], Nlev[pos_i], Nlev[pos_j],
                   Nseq, jntprobIJ);
        }


        /* calc. joint frequencies of positions j and k */
        freq2D(x[pos_j], x[pos_k], Nlev[pos_j], Nlev[pos_k],
               Nseq, jntprobJK);

        /* calc. joint frequencies of positions i and k */
        freq2D(x[pos_i], x[pos_k], Nlev[pos_i], Nlev[pos_k],
               Nseq, jntprobIK);

        /* calc. joint frequencies of positions i, j, and k */
        freq3D(x[pos_i], x[pos_j], x[pos_k], Nlev[pos_i], Nlev[pos_j],
               Nlev[pos_k], Nseq, jntprobIJK);


        tmpVal = 0; 
        for (i_lev = 0; i_lev < Nlev[pos_i]; i_lev++)
        {
            for (j_lev = 0; j_lev < Nlev[pos_j]; j_lev++)
            {
                for (k_lev = 0; k_lev < Nlev[pos_k]; k_lev++)
                {
                    pIJK = jntprobIJK[i_lev][j_lev][k_lev];
                    if ( pIJK > 0 ) 
                    {
                        prod1Dmarg = p[pos_i][i_lev] * p[pos_j][j_lev] *
                                       p[pos_k][k_lev];

                        prod2Dmarg = jntprobIJ[i_lev][j_lev] * 
                                       jntprobJK[j_lev][k_lev] * 
                                       jntprobIK[i_lev][k_lev]; 

                        tmpVal += pIJK * log(pIJK * prod1Dmarg / prod2Dmarg);
                    }
                }
            }
        }

        /* If norm_const=0 then Mstar will be 0/0 = NaN.  When *
         * this occurs we note it by setting *anyNAN = TRUE    */
        norm_const = sum_PlogP[pos_i] + sum_PlogP[pos_j] + sum_PlogP[pos_k];

        M[triple] = tmpVal; 
        Mstar[triple] = -3 * tmpVal / norm_const;

        /* Note: values of 'norm_const' are always <= zero */ 
        if (norm_const == 0) 
            *anyNAN = TRUE;

    } /* end triple loop */
}

/* End calcStatsTriples */


/* =========================================================================== */


void singleGroupTriples(int *xvec, int *ptNseq, int *ptNpos, int *Nlev,
                        int *posI, int *posJ, int *posK, 
                        int *ptNtriples, double *pvec, int *ptNperm,
                        double *M, double *M_p, double *Mstar, double *Mstar_p)
{

    /* define variable types */
    int i, j, k, triple;
    int Nseq, Npos, Ntriples, Nperm;
    long double tmpVal;
     
    Nseq = *ptNseq;
    Npos = *ptNpos;
    Ntriples = *ptNtriples;
    Nperm = *ptNperm;


    /*** BEGIN HOUSEKEEPING ***/

    /* All matrices/list from R have to be passed to C as vectors, 
     * but for clarity, we'd like to access them as if they are matrices.
     * So we set up the infra-structure to allow us to do that. 
     */

    /* Create int matrix 'x' of dim Npos x Nseq to access data in 'xvec' */
    int **x;
    x = Calloc(Npos, int *);
    for (i = 0; i < Npos; i++)
         x[i] = & ( xvec[i*Nseq] );

    /* Create 'p', length Npos, to access marginal prob.s of each position */
    double **p;
    int loc;
    p = Calloc(Npos, double *);
    loc = 0;
    for (i = 0; i < Npos; i++)
    {
         p[i] = & ( pvec[loc] );
	 loc += Nlev[i];
    }



    /* Find the maximum number of levels across all positions: maxLvl */
    int maxLvl;
    maxLvl = 0;
    for (i = 0; i < Npos; i++)
	if (Nlev[i] > maxLvl)
            maxLvl = Nlev[i]; 

    /* Create three 2d arrays of dimension (maxLvl x maxLvl) to be used
     * within 'calcStatsTriples' to store joint freq.s.  Also create a 3d array
     * of dim (maxLvl x maxLvl x maxLvl) for storing 3d joint freq.s  
     */ 
    double **jntprobIJ, **jntprobJK, **jntprobIK; 
    jntprobIJ = Calloc(maxLvl, double *);
    jntprobJK = Calloc(maxLvl, double *);
    jntprobIK = Calloc(maxLvl, double *);

    double ***jntprobIJK; 
    jntprobIJK = Calloc(maxLvl, double **);

    for (i = 0; i < maxLvl; i++)
    {
        jntprobIJ[i] = Calloc(maxLvl, double);
        jntprobJK[i] = Calloc(maxLvl, double);
        jntprobIK[i] = Calloc(maxLvl, double);

        jntprobIJK[i] = Calloc(maxLvl, double *);
        for (j = 0; j < maxLvl; j++)
            jntprobIJK[i][j] = Calloc(maxLvl, double);
    }



    /* Pre-calculate terms needed for statistic Mstar */
    double *sum_PlogP;
    sum_PlogP = Calloc(Npos, double);
    for (i = 0; i < Npos; i++)
    {
       tmpVal = 0; 
       for (k = 0; k < Nlev[i]; k++)
           tmpVal += p[i][k] * log( p[i][k] );

       sum_PlogP[i] = tmpVal; 
    }

    /*** HOUSEKEEPING DONE ***/



    /*** NOW LETS DO SOME WORK! ***/

    /* define/initialize indicator of NAN occurrance in calcStatsTriplet() */
    int anyNAN;
    anyNAN = FALSE;

    /* (1) calculate statistics using the original data */
    calcStatsTriples(x, Nseq, Npos, Ntriples, posI, posJ, posK, Nlev,
              maxLvl, jntprobIJ, jntprobJK, jntprobIK, jntprobIJK,
	      p, sum_PlogP, M, Mstar, &anyNAN);

    /* This code should never run as original data was screened to remove *
     * remove positions with no variation in them. But just in case ...   */
    if (anyNAN == TRUE)
    {
        printf("%s", "Encountered unexpected problem w/original data:\n");
        printf("%s", "NaN values produced for Mstar in function",
		     " calcStatsTriples.  Exiting...\n\n");
        exit(123);
    }


    /* (2) Define/allocate/initialize storage for null data, null statistics,
     *     and for counts of null statistics more extreme than observed ones
     */ 
    int **null_x;
    null_x = Calloc(Npos, int *);
    for (i=0; i<Npos; i++) 
        null_x[i] = Calloc(Nseq, int);

    double *null_M, *null_Mstar;
    null_M = Calloc(Ntriples, double);
    null_Mstar = Calloc(Ntriples, double);

    int *null_M_cnt, *null_Mstar_cnt;
    null_M_cnt = Calloc(Ntriples, int);
    null_Mstar_cnt = Calloc(Ntriples, int);

    for (triple = 0; triple < Ntriples; triple++)
    {
        null_M_cnt[triple] = 0;
        null_Mstar_cnt[triple] = 0; 
    }


    /* More misc. storage allocation, etc */
    int *indx;
    indx = Calloc(Nseq, int);

    double *runifVec;
    runifVec = Calloc(Nseq, double);


    /* Initialize R's RNG (Random Number Generator) */
    GetRNGstate();


    /* (3) Repeatedly create "null" data sets, compute statistics based
     *     upon them, and record whether the null stats are more "extreme"
     *     than the stats calculated from the observed data 
     */
    int iter;
    for (iter=0; iter<Nperm; iter++)
    {

        /* reorder each row of 'x' separately */
	for (i=0; i<Npos; i++) 
        {
            /* generate unif(0,1) vals and initialize indx to 0, 1, .., n-1 */
            for (j=0; j < Nseq; j++)
	    { 
                runifVec[j] = runif(0,1);
                indx[j] = j;
            }

            /* get indx ordering the random numbers (i.e. random index) */
            indexSort(runifVec, indx, Nseq);

            /* Re-order i-th row of 'x' */
            for (j=0; j < Nseq; j++)
                null_x[i][j] = x[i][indx[j]];
        }

        /* calc statistics on current null data set */
        calcStatsTriples(null_x, Nseq, Npos, Ntriples, posI, posJ, posK,
                         Nlev, maxLvl, jntprobIJ, jntprobJK, jntprobIK,
                         jntprobIJK, p, sum_PlogP, null_M, null_Mstar, &anyNAN);

        /* Keep counts of more extreme values */ 
        for (triple = 0; triple < Ntriples; triple++)
        {
            if (null_M[triple] >= M[triple] - DBL_EPSILON)
                null_M_cnt[triple]++ ;

            if (null_Mstar[triple] >= Mstar[triple] - DBL_EPSILON)
                null_Mstar_cnt[triple]++ ; 
        }
    }  
    /* End of permutation loop */


    /* Since the single group analysis uses permutation to estimate p-vals, *
     * we should never get NaNs (since we didn't for the original data,     *
     * else we wouldn't have made it to here)  But leave in as a check      */
    if (anyNAN == TRUE)
    {
        printf("%s", "Encountered unexpected problem w/permuted data:\n");
        printf("%s", "NaN values produced for Mstar in function",
		     " calcStatsTriples.  Exiting...\n\n");
        exit(456);
    }


    /*   Calculate raw (unadjusted) pvals for statistics.
     * Pvals computed including the observed value as on of the null  *
     * values.  Note also that the counts of more extreme values are  *
     * one-sided, so we need to double them to get a two-sided pvalue */
    double Nplus1;
    Nplus1 = Nperm + 1;
    for (triple = 0; triple < Ntriples; triple++)
    {
        /* "Better" computation of raw p-values - includes observed count */
        M_p[triple]     = fmin(1, (2*null_M_cnt[triple] + 1)/Nplus1);
        Mstar_p[triple] = fmin(1, (2*null_Mstar_cnt[triple] + 1)/Nplus1);
    }



    /************** Free all memory allocated with Calloc *************/

    free(x);
    free(p); 
    free(sum_PlogP);
    free(null_M);
    free(null_Mstar);
    free(null_M_cnt);
    free(null_Mstar_cnt);
    free(indx);
    free(runifVec);

    for (k=0; k<Npos; k++) 
        free(null_x[k]);
    free(null_x); 

    for (i = 0; i < maxLvl; i++) 
    {
        free(jntprobIJ[i]);
        free(jntprobJK[i]);
        free(jntprobIK[i]);

	for (j = 0; j < maxLvl; j++) 
            free(jntprobIJK[i][j]);

        free(jntprobIJK[i]);
    }
    free(jntprobIJ);
    free(jntprobJK);
    free(jntprobIK);
    free(jntprobIJK);

    /*********** End freeing memory ***************/ 


    /* Copy RNG state back to R */
    PutRNGstate();

    /** We're done! YAY! **/
}  
/* end 'singleGroupTriples' */



/* =========================================================================== */



void compareGroupsTriples(int *xvec, int *yvec, int *ptNseq_x, int *ptNseq_y,
                          int *ptNpos, int *Nlev_x, int *Nlev_y,
                          int *posI, int *posJ, int *posK, int *ptNtriples,
                          double *xpvec, double *ypvec, int *ptNboot,
			  int *usableNullSampsCnt,
                          double *M, double *M_p, double *Mstar, double *Mstar_p)
{

/** Looking like we won't need xpvec, and ypvec, well, we can use them to 
    calc the observed data stats, then perhaps use them to store the 
    bootstrap stats too...,  same for Nlev_x and Nlev_y
 **/

    /* define variable types */
    int i, j, k, triple;
    int Nseq_x, Nseq_y, Npos, Ntriples, useWt, Nboot;
    long double tmpVal;
    
     
    Nseq_x = *ptNseq_x;
    Nseq_y = *ptNseq_y;
    Npos = *ptNpos;
    Ntriples = *ptNtriples;
    Nboot = *ptNboot;


    /*** BEGIN HOUSEKEEPING ***/

    /* All matrices/list from R have to be passed to C as vectors,
     * but for clarity, we'd like to access them as if they are matrices.
     * So we set up the infra-structure to allow us to do that.
     */

    /* Create int matrix 'x' of dim Npos x Nseq_x to access data in 'xvec' */
    /* and int matrix 'y' of dim Npos x Nseq_y to access data in 'yvec' */
    int **x, **y;
    x = Calloc(Npos, int *);
    y = Calloc(Npos, int *);
    for (i = 0; i < Npos; i++) {
         x[i] = & ( xvec[ i * Nseq_x ] );
         y[i] = & ( yvec[ i * Nseq_y ] );
    }


    /* Create 'p_x' and 'p_y', each length Npos, to access marginal prob.s of */
    /* of groups 'x' and 'y' at each position of the seqs*/
    double **p_x, **p_y;
    int loc_x, loc_y;

    p_x = Calloc(Npos, double *);
    p_y = Calloc(Npos, double *);

    loc_x = 0;
    loc_y = 0;

    for (i = 0; i < Npos; i++)
    {
         p_x[i] = & ( xpvec[loc_x] );
         loc_x += Nlev_x[i];

         p_y[i] = & ( ypvec[loc_y] );
         loc_y += Nlev_y[i];
    }



    /* Next, we set up a matrix to store cross-tab results for a pair of pos.,
     * then pre-calculate some quantities needed for statistic Mstar.
     */

    /* Find the maximum number of levels across all positions and both x and y */
    int maxLvl;
    maxLvl = 0;
    for (i = 0; i < Npos; i++) 
    {
        if (Nlev_x[i] > maxLvl)
            maxLvl = Nlev_x[i];
        if (Nlev_y[i] > maxLvl)
            maxLvl = Nlev_y[i];
    } 


    /* Create three 2d arrays of dimension (maxLvl x maxLvl) to be used
     * within 'calcStatsTriples' to store joint freq.s.  Also create a 3d array
     * of dim (maxLvl x maxLvl x maxLvl) for storing 3d joint freq.s  
     */ 
    double **jntprobIJ, **jntprobJK, **jntprobIK; 
    jntprobIJ = Calloc(maxLvl, double *);
    jntprobJK = Calloc(maxLvl, double *);
    jntprobIK = Calloc(maxLvl, double *);

    double ***jntprobIJK; 
    jntprobIJK = Calloc(maxLvl, double **);

    for (i = 0; i < maxLvl; i++)
    {
        jntprobIJ[i] = Calloc(maxLvl, double);
        jntprobJK[i] = Calloc(maxLvl, double);
        jntprobIK[i] = Calloc(maxLvl, double);

        jntprobIJK[i] = Calloc(maxLvl, double *);
        for (j = 0; j < maxLvl; j++)
            jntprobIJK[i][j] = Calloc(maxLvl, double);
    }



    /* Pre-calculate terms needed for statistic Mstar 
     */
    double *sum_PlogP_x,  *sum_PlogP_y;
    sum_PlogP_x = Calloc(Npos, double);
    sum_PlogP_y = Calloc(Npos, double);

    for (i = 0; i < Npos; i++)
    {
       tmpVal = 0;
       for (k = 0; k < Nlev_x[i]; k++)
           tmpVal += p_x[i][k] * log( p_x[i][k] );

       sum_PlogP_x[i] = tmpVal;

       tmpVal = 0;
       for (k = 0; k < Nlev_y[i]; k++)
           tmpVal += p_y[i][k] * log( p_y[i][k] );

       sum_PlogP_y[i] = tmpVal;
    }

    /* Storage for individual group statistics */
    double *M_x, *M_y, *Mstar_x, *Mstar_y;
    M_x = Calloc(Ntriples, double);
    M_y = Calloc(Ntriples, double);
    Mstar_x = Calloc(Ntriples, double);
    Mstar_y = Calloc(Ntriples, double);

    /* Storage for sign of statistics */
    int *sign_M, *sign_Mstar;
    sign_M = Calloc(Ntriples, int);
    sign_Mstar = Calloc(Ntriples, int);

    /*** HOUSEKEEPING DONE ***/




    /*** NOW LETS DO SOME WORK! ***/

    /* define/initialize indicator of NAN occurrance in calcStatsTriplet() */
    int anyNAN;
    anyNAN = FALSE;

    /* (1) calculate statistics using the original data */
    calcStatsTriples(x, Nseq_x, Npos, Ntriples, posI, posJ, posK, 
                     Nlev_x, maxLvl, jntprobIJ, jntprobJK, jntprobIK,
                     jntprobIJK, p_x, sum_PlogP_x, M_x, Mstar_x, &anyNAN);

    calcStatsTriples(y, Nseq_y, Npos, Ntriples, posI, posJ, posK, 
                     Nlev_y, maxLvl, jntprobIJ, jntprobJK, jntprobIK,
                     jntprobIJK, p_y, sum_PlogP_y, M_y, Mstar_y, &anyNAN);


    /* This code should never run as original data was screened to remove *
     * remove positions with no variation in them. But just in case ...   */
    if (anyNAN == TRUE)
    {
        printf("%s", "Encountered unexpected problem w/original data:\n");
        printf("%s", "NaN values produced for Mstar in function",
		     " calcStatsTriples.  Exiting...\n\n");
        exit(123);
    }


    /* calculate difference statistics: just difference of the stats */
    for (triple=0; triple < Ntriples; triple++)
    {
        M[triple] = M_x[triple] - M_y[triple];
        Mstar[triple] = Mstar_x[triple] - Mstar_y[triple];

        sign_M[triple] = Sign(M[triple]);
        sign_Mstar[triple] = Sign(Mstar[triple]);
    }



    /* (2) Define/allocate/initialize storage for null data, null statistics,
     *     and for counts of null statistics more extreme than observed ones
     */
    int **x_boot, **y_boot;
    x_boot = Calloc(Npos, int *);
    y_boot = Calloc(Npos, int *);
    for (i=0; i<Npos; i++)
     {
        x_boot[i] = Calloc(Nseq_x, int);
        y_boot[i] = Calloc(Nseq_y, int);
     }

    /* More misc. storage allocation, etc */
    int *x_boot_indx, *y_boot_indx;
    x_boot_indx = Calloc(Nseq_x, int);
    y_boot_indx = Calloc(Nseq_y, int);

    double *boot_M, *boot_Mstar, *boot_M_x, *boot_M_y,
           *boot_Mstar_x, *boot_Mstar_y;
    boot_M = Calloc(Ntriples, double);
    boot_Mstar = Calloc(Ntriples, double);
    boot_M_x = Calloc(Ntriples, double);
    boot_M_y = Calloc(Ntriples, double);
    boot_Mstar_x = Calloc(Ntriples, double);
    boot_Mstar_y = Calloc(Ntriples, double);

    /* counts of number of NaN values produced - these are not used */
    int *boot_NANcnt;
    boot_NANcnt = Calloc(Ntriples, int);

    /* counts of bootstrapped stats "on the other side of zero" from stats*/
    int *boot_M_cnt, *boot_Mstar_cnt;
    boot_M_cnt = Calloc(Ntriples, int);
    boot_Mstar_cnt = Calloc(Ntriples, int);

    for (triple=0; triple < Ntriples; triple++)
    {
        boot_NANcnt[triple] = 0;
        boot_M_cnt[triple] = 0;
        boot_Mstar_cnt[triple] = 0;
    }



    /* Initialize R's RNG (Random Number Generator) */
    GetRNGstate();


    /* (3) Repeatedly bootstrap the data sets, compute statistics based
     *     upon them, and record whether the stats ... 
     */
    int iter;
    for (iter=0; iter<Nboot; iter++)
    {

        /* get indices of seqs that comprise the current bootstrap sample */
        /* Values are in the set {0, 1,..., Nseq -1 } */ 
        for (j = 0; j < Nseq_x; j++) 
           x_boot_indx[j] = floor( Nseq_x * runif(0,1) );

        for (j = 0; j < Nseq_y; j++) 
           y_boot_indx[j] = floor( Nseq_y * runif(0,1) );


        /* create bootstrap data sets*/
        for (i=0; i<Npos; i++)
        {
            for(j=0; j<Nseq_x; j++) 
                x_boot[i][j] = x[i][ x_boot_indx[j] ];

            for(j=0; j<Nseq_y; j++) 
                y_boot[i][j] = y[i][ y_boot_indx[j] ]; 
        } 


        /* Calculate vectors p for 'x_boot' and 'y_boot' */
        for (i = 0; i < Npos; i++)
        {
            freq1D(x[i], Nlev_x[i], Nseq_x, p_x[i]);
            freq1D(y[i], Nlev_y[i], Nseq_y, p_y[i]);
        }

        /* Pre-calculate terms needed for statistic Mstar */
        /* Note - we are re-using the same objects used for original data values */
        double p_ik;
        for (i = 0; i < Npos; i++)
        {
           tmpVal = 0;
           for (k = 0; k < Nlev_x[i]; k++) 
           {
               p_ik = p_x[i][k]; 
               if ( p_ik  > 0 )
                   tmpVal += p_ik * log( p_ik );
           }
           sum_PlogP_x[i] = tmpVal;


           tmpVal = 0;
           for (k = 0; k < Nlev_y[i]; k++)
           {
               p_ik = p_y[i][k]; 
               if ( p_ik  > 0 )
                   tmpVal += p_ik * log( p_ik );
           }
           sum_PlogP_y[i] = tmpVal;
        }


        /* calc stats from bootstrapped data sets */
        calcStatsTriples(x_boot, Nseq_x, Npos, Ntriples, posI, posJ, posK,
                         Nlev_x, maxLvl, jntprobIJ, jntprobJK, jntprobIK,
                         jntprobIJK, p_x, sum_PlogP_x, boot_M_x, boot_Mstar_x,
                         &anyNAN);

        calcStatsTriples(y_boot, Nseq_y, Npos, Ntriples, posI, posJ, posK, 
                         Nlev_y, maxLvl, jntprobIJ, jntprobJK, jntprobIK,
                         jntprobIJK, p_y, sum_PlogP_y, boot_M_y, boot_Mstar_y,
                         &anyNAN);


       /* calc difference in stats between the two groups */
        double bM, bMs;
        for (triple=0; triple < Ntriples; triple++)
        {
            bM  = boot_M_x[triple] - boot_M_y[triple];
            bMs = boot_Mstar_x[triple] - boot_Mstar_y[triple];

            boot_M[triple] = bM;
            boot_Mstar[triple] = bMs;

            /* count bootstrapped stats on "other side of zero" from stats */
            if (anyNAN == FALSE || !isnan(bMs))
            {
                if ( Sign(bM) != sign_M[triple] || fabs(bM) < DBL_EPSILON)
                    boot_M_cnt[triple]++ ;

                if ( Sign(bMs) != sign_Mstar[triple] || fabs(bMs) < DBL_EPSILON)
                    boot_Mstar_cnt[triple]++ ;
	    } else
            {
                boot_NANcnt[ triple ]++ ;
            }
        }

        /* reset NAN indicator for next bootstrap sample */
        if (anyNAN == TRUE)
            anyNAN = FALSE;

    }
    /* End bootstrap loop */



    /*   Calculate raw (unadjusted) pvals for statistics. */
    double nonNANcnt_plus1;
    for (triple = 0; triple < Ntriples; triple++)
    {
        /* Pvals computed including the observed value as on of the null  *
         * values.  Note also that the counts of more extreme values are  *
         * one-sided, so we need to double them to get a two-sided pvalue */
        usableNullSampsCnt[triple] =  Nboot - boot_NANcnt[triple];
        nonNANcnt_plus1 = usableNullSampsCnt[triple] + 1;

        M_p[triple]     = fmin(1, (2*boot_M_cnt[triple] + 1)/nonNANcnt_plus1) ;
        Mstar_p[triple] = fmin(1, (2*boot_Mstar_cnt[triple] + 1)/nonNANcnt_plus1);
    }



    /* Free all memory allocated with Calloc */
    free(x);
    free(y);
    free(p_x); 
    free(p_y); 
    free(sum_PlogP_x);
    free(sum_PlogP_y);
    free(x_boot_indx);
    free(y_boot_indx);
    free(M_x);
    free(M_y);
    free(Mstar_x);
    free(Mstar_y);
    free(sign_M);
    free(sign_Mstar);

    free(boot_M);
    free(boot_Mstar);

    free(boot_M_x);
    free(boot_Mstar_x);
    free(boot_M_y);
    free(boot_Mstar_y);
    free(boot_M_cnt);
    free(boot_Mstar_cnt);
    free(boot_NANcnt);

    for (k=0; k<Npos; k++) 
    {
        free(x_boot[k]);
        free(y_boot[k]);
    }
    free(x_boot); 
    free(y_boot); 

    for (i = 0; i < maxLvl; i++) 
    {
        free(jntprobIJ[i]);
        free(jntprobJK[i]);
        free(jntprobIK[i]);

	for (j = 0; j < maxLvl; j++) 
            free(jntprobIJK[i][j]);

        free(jntprobIJK[i]);
    }
    free(jntprobIJ);
    free(jntprobJK);
    free(jntprobIK);
    free(jntprobIJK);

    /*********** End freeing memory ***************/ 


    /* Copy RNG state back to R */
    PutRNGstate();

}
/* end 'compareGroupsTriples' */

